import ssl
from typing import Awaitable, Callable, Literal, TypeAlias

SSL_CONTEXT = ssl._create_unverified_context()
DEFAULT_REST_API_HOST = "api-global-prod.aircloudhome.com"
DEFAULT_STOMP_WEBSOCKET_HOST = "notification-global-prod.aircloudhome.com"

TokenSupplier: TypeAlias = Callable[[], Awaitable[str]]

ApiCommandState: TypeAlias = Literal["SENDING", "INCOMPLETE", "DONE"]

FanSwing: TypeAlias = Literal["OFF", "VERTICAL", "HORIZONTAL", "BOTH", "AUTO"]
FanSpeed: TypeAlias = Literal["LV1", "LV2", "LV3", "LV4", "LV5", "AUTO"]
Power: TypeAlias = Literal["ON", "OFF"]

OperatingMode: TypeAlias = Literal["AUTO", "COOLING", "DRY", "FAN", "HEATING"]
TemperatureUnit: TypeAlias = Literal["CELSIUS", "FAHRENHEIT"]

ScheduleType: TypeAlias = Literal[
    "SCHEDULE_DISABLED",
    "OFF_TIMER_ENABLED",
    "ON_TIMER_ENABLED",
    "ON_OFF_TIMER_ENABLED",
    "WEEKLY_TIMER_ENABLED",
    "HOLIDAY_MODE_ENABLED",
]



import datetime

from .api.rac_models import InteriorUnitUserState
from .contants import FanSpeed, FanSwing, OperatingMode, Power, ScheduleType


class InteriorUnitBase:
    _rac_id: int
    _name: str
    _room_temperature: float
    _relative_temperature: float
    _updated_at: datetime.datetime
    _online: bool
    _online_updated_at: datetime.datetime
    _vendor: str
    _model_id: str
    _serial_number: str
    _vendor_thing_id: str
    _schedule_type: ScheduleType
    _user_state: InteriorUnitUserState

    def __init__(
        self,
        rac_id: int,
        name: str,
        room_temperature: float,
        relative_temperature: float,
        updated_at: datetime.datetime,
        online: bool,
        online_updated_at: datetime.datetime,
        vendor: str,
        model_id: str,
        serial_number: str,
        vendor_thing_id: str,
        schedule_type: ScheduleType,
        power: Power,
        operating_mode: OperatingMode,
        requested_temperature: float,
        humidity: int,
        fan_speed: FanSpeed,
        fan_swing: FanSwing,
    ) -> None:
        self._rac_id = rac_id
        self._name = name
        self._room_temperature = room_temperature
        self._relative_temperature = relative_temperature
        self._updated_at = updated_at
        self._online = online
        self._online_updated_at = online_updated_at
        self._vendor = vendor
        self._model_id = model_id
        self._serial_number = serial_number
        self._vendor_thing_id = vendor_thing_id
        self._schedule_type = schedule_type
        self._user_state = InteriorUnitUserState(
            rac_id, power, operating_mode, requested_temperature, humidity, fan_speed, fan_swing
        )

    @property
    def rac_id(self) -> int:
        return self._rac_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def room_temperature(self) -> float:
        return self._room_temperature

    @property
    def relative_temperature(self) -> float:
        return self._relative_temperature

    @property
    def updated_at(self) -> datetime.datetime:
        return self._updated_at

    @property
    def online(self) -> bool:
        return self._online

    @property
    def online_updated_at(self) -> datetime.datetime:
        return self._online_updated_at

    @property
    def vendor(self) -> str:
        return self._vendor

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def user_state(self) -> InteriorUnitUserState:
        return self._user_state

    @property
    def power(self) -> Power:
        return self._user_state._power

    @property
    def operating_mode(self) -> OperatingMode:
        return self._user_state._operating_mode

    @property
    def requested_temperature(self) -> float:
        return self._user_state._requested_temperature

    @property
    def humidity(self) -> int:
        return self._user_state._humidity

    @property
    def fan_speed(self) -> FanSpeed:
        return self._user_state._fan_speed

    @property
    def fan_swing(self) -> FanSwing:
        return self._user_state._fan_swing



from __future__ import annotations

from dataclasses import dataclass

from aircloudy.contants import FanSpeed, FanSwing, OperatingMode, Power


@dataclass
class CommandResponse:
    commandId: str
    thingId: str

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


@dataclass
class PowerResult:
    racId: int
    success: bool
    errorMessage: str | None
    errorCode: int
    commandResponse: CommandResponse

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


@dataclass
class PowerAllResponse:
    allSucceeded: bool
    resultSet: list[PowerResult]

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


class InteriorUnitUserState:
    _rac_id: int
    _power: Power
    _operating_mode: OperatingMode
    _requested_temperature: float
    _humidity: int
    _fan_speed: FanSpeed
    _fan_swing: FanSwing

    def __init__(
        self,
        rac_id: int,
        power: Power,
        operating_mode: OperatingMode,
        requested_temperature: float,
        humidity: int,
        fan_speed: FanSpeed,
        fan_swing: FanSwing,
    ) -> None:
        self._rac_id = rac_id
        self._power = power
        self._operating_mode = operating_mode
        self._requested_temperature = requested_temperature
        self._humidity = humidity
        self._fan_speed = fan_speed
        self._fan_swing = fan_swing

    @property
    def rac_id(self) -> int:
        return self._rac_id

    @property
    def power(self) -> Power:
        return self._power

    @property
    def operating_mode(self) -> OperatingMode:
        return self._operating_mode

    @property
    def requested_temperature(self) -> float:
        return self._requested_temperature

    @property
    def humidity(self) -> int:
        return self._humidity

    @property
    def fan_speed(self) -> FanSpeed:
        return self._fan_speed

    @property
    def fan_swing(self) -> FanSwing:
        return self._fan_swing

    def copy(
        self,
        power: Power | None = None,
        mode: OperatingMode | None = None,
        requested_temperature: float | None = None,
        humidity: int | None = None,
        fan_speed: FanSpeed | None = None,
        fan_swing: FanSwing | None = None,
    ) -> InteriorUnitUserState:
        return InteriorUnitUserState(
            self._rac_id,
            self._power if power is None else power,
            self._operating_mode if mode is None else mode,
            self._requested_temperature if requested_temperature is None else requested_temperature,
            self._humidity if humidity is None else humidity,
            self._fan_speed if fan_speed is None else fan_speed,
            self._fan_swing if fan_swing is None else fan_swing,
        )

    def to_api(self) -> dict:
        return {
            "id": self._rac_id,
            "power": self._power,
            "mode": self._operating_mode,
            "iduTemperature": self._requested_temperature,
            # Should be something like `self._humidity`
            # but some api return me humidity=126 and this is not a legal value for control
            "humidity": 50,
            "fanSpeed": self.fan_speed,
            "fanSwing": self.fan_swing,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InteriorUnitUserState):
            return False

        return self.__dict__ == other.__dict__



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, Self

from . import api, notifications
from .api.rac_models import InteriorUnitUserState
from .contants import (
    DEFAULT_REST_API_HOST,
    DEFAULT_STOMP_WEBSOCKET_HOST,
    TemperatureUnit,
)
from .errors import IllegalStateException, InteriorUnitNotFoundException
from .interior_unit import InteriorUnit
from .interior_unit_base import InteriorUnitBase
from .interior_unit_changes import InteriorUnitChanges

logger = logging.getLogger(__name__)


@dataclass
class ConnectionInfo:
    auth_manager: api.AuthManager
    user_profile: api.UserProfile
    notification_socket: notifications.NotificationsWebsocket


class HitachiAirCloud:
    _email: str
    _password: str
    _api_host: str
    _api_port: int

    _command_state_monitor: api.CommandStateMonitor
    _connection_info: ConnectionInfo | None
    _interior_units: dict[int, InteriorUnit]

    on_change: Callable[[dict[int, InteriorUnitChanges]], None] | None

    def __init__(
        self,
        email: str,
        password: str,
        api_host: str = DEFAULT_REST_API_HOST,
        api_port: int = 443,
        notification_host: str = DEFAULT_STOMP_WEBSOCKET_HOST,
    ) -> None:
        self._email = email
        self._password = password
        self._api_host = api_host
        self._api_port = api_port
        self.notification_host = notification_host

        self._command_state_monitor = api.CommandStateMonitor(
            self._get_auth_token_or_fail, host=api_host, port=api_port
        )
        self._connection_info = None
        self._interior_units = {}

        self.on_change = None

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        await self.close()
        return False

    @property
    def is_open(self) -> bool:
        return self._connection_info is not None

    @property
    def interior_units(self) -> list[InteriorUnit]:
        return list(self._interior_units.values())

    @property
    def temperature_unit(self) -> TemperatureUnit:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")

        if self._connection_info.user_profile.settings.temperatureUnit == "degC":
            return "CELSIUS"

        return "FAHRENHEIT"

    def find_interior_unit(self, rac_id: int) -> InteriorUnit | None:
        return self._interior_units.get(rac_id)

    def get_interior_unit(self, rac_id: int) -> InteriorUnit:
        iu = self.find_interior_unit(rac_id)
        if iu is None:
            raise InteriorUnitNotFoundException(f"Interior unit {rac_id} not found")
        return iu

    async def _get_auth_token_or_fail(self) -> str:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")
        return await self._connection_info.auth_manager.token()

    async def connect(self) -> None:
        if self.is_open:
            raise IllegalStateException("AirCloud already connected")

        auth_manager = api.AuthManager(self._email, self._password, self._api_host, self._api_port)
        user_profile = await api.fetch_profile(auth_manager.token, self._api_host, self._api_port)
        self._interior_units = {
            iu.rac_id: InteriorUnit(self._send_command_and_wait_ack, iu)
            for iu in await api.get_interior_units(
                auth_manager.token, user_profile.familyId, self._api_host, self._api_port
            )
        }

        notification_socket = notifications.NotificationsWebsocket(
            self.notification_host,
            auth_manager.token,
            user_profile.id,
            user_profile.familyId,
            self._update_interior_units,
        )
        notification_socket.on_unexpected_connection_close = lambda _: self._init_notification_socket(
            notification_socket
        )
        await self._init_notification_socket(notification_socket)

        self._connection_info = ConnectionInfo(
            auth_manager,
            user_profile,
            notification_socket,
        )

        logger.info("Connected")

    async def _init_notification_socket(self, socket: notifications.NotificationsWebsocket) -> None:
        await socket.connect()
        await socket.subscribe()

    async def close(self) -> None:
        try:
            if self._connection_info is not None:
                await self._connection_info.notification_socket.close()
        finally:
            self._connection_info = None
            self._interior_units = {}
        logging.info("Closed")

    def _update_interior_units(self, interior_units: list[InteriorUnitBase], partial: bool) -> None:
        logger.debug("Received interior units update: %s", interior_units)
        changes: dict[int, InteriorUnitChanges] = {}
        for iu in interior_units:
            changes[iu.rac_id] = self._interior_units[iu.rac_id].update(iu)

        # If update is not partial, compare given list and current list to detected deleted interior_unit
        # and notify change
        if partial:
            pass

        if self.on_change is not None:
            self.on_change(changes)

    async def update_all(self) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        self._update_interior_units(
            await api.get_interior_units(
                self._connection_info.auth_manager.token,
                self._connection_info.user_profile.familyId,
                self._api_host,
                self._api_port,
            ),
            False,
        )

    async def request_update_all(self) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        await self._connection_info.notification_socket.refresh_all()

    async def request_update(self, rac_id: int) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        await self._connection_info.notification_socket.refresh(rac_id)

    async def _send_command_and_wait_ack(
        self,
        interior_unit_command: InteriorUnitUserState,
    ) -> None:
        """
        Send command to set interior_unit state

        :raises:
            IllegalStateException: If instance is not connected
            CommandFailedException: If state isn't done after wait delay
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        command_state = await self._command_state_monitor.watch_command(
            await api.send_command(
                self._connection_info.auth_manager.token,
                self._connection_info.user_profile.familyId,
                interior_unit_command,
                host=self._api_host,
                port=self._api_port,
            )
        )
        await asyncio.wait_for(command_state.wait_done(), 30)
        await self.request_update(interior_unit_command.rac_id)



from .auth_manager import AuthManager
from .command_state_monitor import CommandStateMonitor
from .iam import fetch_profile, perform_login
from .iam_models import AuthenticationSuccess, UserProfile
from .rac import (
    get_commands_state,
    get_interior_units,
    request_refresh_interior_unit_state,
    send_command,
    set_power,
    set_power_all,
)
from .rac_models import CommandResponse



from __future__ import annotations

import asyncio
import logging
import traceback
import uuid
from asyncio import Task
from types import TracebackType
from typing import Awaitable, Callable, Self

import websockets

from aircloudy.contants import SSL_CONTEXT, TokenSupplier
from aircloudy.errors import IllegalStateException
from aircloudy.utils import current_task_is_running, utc_datetime_from_millis

from ..interior_unit_base import InteriorUnitBase
from . import hitachi_frame_models, stomp

logger = logging.getLogger(__name__)


class NotificationsWebsocket:
    _notification_host: str
    _token_supplier: TokenSupplier
    _user_id: int
    _family_id: int
    state_callback: Callable[[list[InteriorUnitBase], bool], None]
    on_unexpected_connection_close: Callable[[websockets.ConnectionClosed], Awaitable[None]] | None

    _notification_socket: websockets.WebSocketClientProtocol | None = None
    _handle_connection_task: Task | None
    _closed_by_client: bool

    def __init__(
        self,
        notification_host: str,
        token_supplier: TokenSupplier,
        user_id: int,
        family_id: int,
        state_callback: Callable[[list[InteriorUnitBase], bool], None],
        on_unexpected_connection_close: Callable[[websockets.ConnectionClosed], Awaitable[None]] | None = None,
    ) -> None:
        self._notification_host = notification_host
        self._token_supplier = token_supplier
        self._user_id = user_id
        self._family_id = family_id
        self.state_callback = state_callback
        self.on_unexpected_connection_close = on_unexpected_connection_close
        self.notification_subscription_id = uuid.uuid4()
        self._closed_by_client = True

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        await self.close()
        return False

    @property
    def is_open(self) -> bool:
        return self._notification_socket is not None

    async def connect(self) -> None:
        self._closed_by_client = False
        await self._init_connection()

        self._handle_connection_task = asyncio.create_task(self._handle_connection())

    async def _handle_connection(self) -> None:
        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self._send_client_heartbeat_loop())
                tg.create_task(self._handle_incoming_frame_loop())
        except ExceptionGroup as e:
            connection_closed: websockets.ConnectionClosed | None = (
                e.exceptions[0]
                if len(e.exceptions) == 1 and isinstance(e.exceptions[0], websockets.ConnectionClosed)
                else None
            )
            if connection_closed is not None:
                logger.info(
                    "Connection closed with status %d and reason %s", connection_closed.code, connection_closed.reason
                )
                if not self._closed_by_client and self.on_unexpected_connection_close is not None:
                    await self.on_unexpected_connection_close(connection_closed)

    async def _init_connection(self) -> None:
        if self.is_open:
            raise Exception(__name__ + " already connected")

        websocket_url = f"wss://{self._notification_host}/rac-notifications/websocket"
        logger.info("Open websocket to %s", websocket_url)
        self._notification_socket = await websockets.connect(websocket_url, ssl=SSL_CONTEXT)
        logger.debug("Send CONNECT stomp frame")
        await self._notification_socket.send(
            hitachi_frame_models.ConnectFrame(await self._token_supplier()).get_frame()
        )

        logger.debug("Wait first server frame (expect CONNECTED frame)")
        frame_data = await self._notification_socket.recv()
        if isinstance(frame_data, bytes):
            raise Exception("Binary frame was unexpected")
        first_server_frame = stomp.parse_server_frame(frame_data)
        if not isinstance(first_server_frame, stomp.ConnectedFrame):
            raise Exception(f"Expected stomp.ConnectedFrame but got {first_server_frame}")

    async def subscribe(self) -> uuid.UUID:
        if self._notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        subscription_id = uuid.uuid4()
        logger.info(
            "Create subscription %s to notifications from user_id=%s, family_id=%s",
            subscription_id,
            self._user_id,
            self._family_id,
        )
        payload = hitachi_frame_models.SubscribeFrame(subscription_id, self._user_id, self._family_id)
        await self._notification_socket.send(payload.get_frame())
        return subscription_id

    async def unsubscribe(self, subscription_id: uuid.UUID) -> None:
        if self._notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Remove subscription %s", subscription_id)
        payload = hitachi_frame_models.UnsubscribeFrame(subscription_id)
        await self._notification_socket.send(payload.get_frame())

    async def refresh_all(self) -> None:
        if self._notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Request refresh all")
        payload = hitachi_frame_models.RefreshAllInteriorUnitFrame(
            await self._token_supplier(), self._user_id, self._family_id
        )
        await self._notification_socket.send(payload.get_frame())

    async def refresh(self, rac_id: int) -> None:
        if self._notification_socket is None:
            raise IllegalStateException(__name__ + " is not connected")

        logger.info("Request refresh rac_id=%d", rac_id)
        payload = hitachi_frame_models.RefreshInteriorUnitFrame(
            await self._token_supplier(), self._user_id, self._family_id, rac_id
        )
        await self._notification_socket.send(payload.get_frame())

    async def _send_client_heartbeat_loop(self) -> None:
        logger.debug("Start send client heartbeat loop")
        while current_task_is_running() and self._notification_socket is not None:
            try:
                logger.debug("Send heartbeat")
                await self._notification_socket.send("\r\n")
                await asyncio.sleep(10)
            except websockets.ConnectionClosed as e:
                raise e
            except Exception as e:
                logger.error("Unexpected error while sending client heartbeat : %s", traceback.format_exc())
                raise e
        logger.debug("End send client heartbeat loop")

    async def _handle_incoming_frame_loop(self) -> None:
        logger.debug("Start handle incoming frame loop")
        while current_task_is_running() and self._notification_socket is not None:
            try:
                logger.debug("Waiting for message")
                data = await self._notification_socket.recv()
                if isinstance(data, bytes):
                    raise Exception("Binary frame was unexpected")

                frame = stomp.parse_server_frame(data)
                logger.debug("Received frame %s", frame)
                match frame:
                    case None:
                        logger.debug("Frame was server-heartbeat")
                    case stomp.ConnectedFrame():
                        raise Exception("ConnectedFrame should have been received at initialization")
                    case stomp.MessageFrame():
                        if frame.body is None:
                            raise Exception("Unexpected message without body")

                        notification_type = frame.body.get("notificationType")
                        if notification_type is None:
                            raise Exception("Unexpected message without notificationType")

                        if notification_type in ("ON_CONNECT", "BUCKET_UPDATE", "REFRESH_ALL"):
                            interior_units = [
                                InteriorUnitBase(
                                    d["id"],
                                    d["name"],
                                    d["roomTemperature"],
                                    d["relativeTemperature"],
                                    utc_datetime_from_millis(d["updatedAt"]),
                                    d["online"],
                                    utc_datetime_from_millis(d["lastOnlineUpdatedAt"]),
                                    d["model"],
                                    str(d["modelTypeId"]),
                                    d["serialNumber"],
                                    d["vendorThingId"],
                                    d["scheduletype"],
                                    d["power"],
                                    d["mode"],
                                    d["iduTemperature"],
                                    d["humidity"],
                                    d["fanSpeed"],
                                    d["fanSwing"],
                                )
                                for d in frame.body["data"]
                            ]
                            self.state_callback(interior_units, notification_type == "BUCKET_UPDATE")
                        else:
                            raise Exception("Unexpected message notification_type", notification_type)

                    case _:
                        logger.warning("Unexpected frame type : %s", frame.message)
            except websockets.ConnectionClosed as e:
                raise e
            except Exception as e:
                logger.error("Unexpected error while handling incoming message : %s", traceback.format_exc())
                raise e
        logger.debug("End handle incoming frame loop")

    async def close(self) -> None:
        self._closed_by_client = True
        try:
            tasks: list[Awaitable] = []
            if self._notification_socket is not None:
                tasks.append(self._notification_socket.close())
            if self._handle_connection_task is not None:
                tasks.append(self._handle_connection_task)

            await asyncio.gather(*tasks)
        finally:
            self._notification_socket = None
        logger.info("Websocket closed")



from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import Literal

import jwt


@dataclass
class UserProfile:
    @dataclass
    class Settings:
        outOfHomeAddress: None
        sensitiveToCold: bool
        temperatureUnit: Literal["degC"]
        outOfHomeLongitude: float
        homeOnWeekdays: bool
        language: Literal["en"]
        outOfHomeRadius: float
        homeOnWeekends: bool
        outOfHomeRemainderEnabled: bool
        outOfHomeLatitude: float

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    @dataclass
    class Address:
        zipCode: str
        city: str
        street: str
        countryCode: str
        state: str
        addressLine: str

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    @dataclass
    class Role:
        level: int
        name: Literal["OWNER"]
        id: int

        def __init__(self, data: dict) -> None:
            self.__dict__.update(data)

    familyId: int
    firstName: str
    lastName: str
    settings: Settings
    address: Address
    phoneNumber: None
    pictureData: None
    familyName: str
    roles: list[Role]
    middleName: None
    id: int
    email: str

    def __init__(self, data: dict) -> None:
        self.__dict__.update(data)


@dataclass
class JWTToken:
    value: str
    sub: str
    scopes: list[str]
    iss: str
    aud: str
    jti: str | None
    iat: datetime.datetime
    exp: datetime.datetime

    def __init__(self, value: str, data: dict) -> None:
        self.value = value
        self.sub = data["sub"]
        self.scopes = data["scopes"]
        self.iss = data["iss"]
        self.aud = data["aud"]
        self.jti = data.get("jti")
        self.iat = datetime.datetime.fromtimestamp(data["iat"], datetime.timezone.utc)
        self.exp = datetime.datetime.fromtimestamp(data["exp"], datetime.timezone.utc)


@dataclass
class AuthenticationSuccess:
    token: JWTToken
    refresh_token: JWTToken
    new_user: bool
    error_state: Literal["NONE"]
    access_token_expires_in: int
    refresh_token_expires_in: int

    def __init__(self, data: dict) -> None:
        self.token = JWTToken(data["token"], jwt.decode(data["token"], options={"verify_signature": False}))
        self.refresh_token = JWTToken(
            data["refreshToken"], jwt.decode(data["refreshToken"], options={"verify_signature": False})
        )
        self.new_user = data["newUser"]
        self.error_state = data["errorState"]
        self.access_token_expires_in = data["access_token_expires_in"]
        self.refresh_token_expires_in = data["refresh_token_expires_in"]


@dataclass
class TokenRefreshSuccess:
    token: JWTToken
    refresh_token: JWTToken
    error_state: Literal["NONE"]
    access_token_expires_in: int

    def __init__(self, data: dict) -> None:
        self.token = JWTToken(data["token"], jwt.decode(data["token"], options={"verify_signature": False}))
        self.refresh_token = JWTToken(
            data["refreshToken"], jwt.decode(data["refreshToken"], options={"verify_signature": False})
        )
        self.error_state = data["errorState"]
        self.access_token_expires_in = data["access_token_expires_in"]



from __future__ import annotations

import asyncio
import datetime

import tzlocal

from aircloudy.contants import DEFAULT_REST_API_HOST
from aircloudy.utils import awaitable

from .iam import perform_login, refresh_token
from .iam_models import JWTToken


class AuthManager:
    _token_update_lock: asyncio.Lock
    _refresh_before_expiration: datetime.timedelta
    _email: str
    _password: str
    _host: str
    _port: int
    _token: JWTToken | None
    _refresh_token: JWTToken | None

    def __init__(self, email: str, password: str, host: str = DEFAULT_REST_API_HOST, port: int = 443) -> None:
        self._token_update_lock = asyncio.Lock()
        self._refresh_before_expiration = datetime.timedelta(minutes=1)
        self._email = email
        self._password = password
        self._host = host
        self._port = port
        self._token = None
        self._refresh_token = None

    async def token(self) -> str:
        async with self._token_update_lock:
            if self._token is None:
                return await self._perform_login()

            now = datetime.datetime.now(tzlocal.get_localzone()).astimezone(datetime.timezone.utc)
            limit_date = self._token.exp - self._refresh_before_expiration
            if now <= limit_date:
                return self._token.value

            if self._refresh_token is not None and self._refresh_token.exp <= limit_date:
                token = awaitable(self._refresh_token.value)
                refresh_result = await refresh_token(lambda: token, self._host, self._port)

                self._token = refresh_result.token
                self._refresh_token = refresh_result.refresh_token
                return refresh_result.token.value

            return await self._perform_login()

    async def _perform_login(self) -> str:
        login_result = await perform_login(self._email, self._password, self._host, self._port)
        self._token = login_result.token
        self._refresh_token = login_result.refresh_token
        return login_result.token.value



from __future__ import annotations

import logging
from typing import Literal

import aiohttp
from aiohttp import TCPConnector

from aircloudy.contants import DEFAULT_REST_API_HOST, SSL_CONTEXT, TokenSupplier
from aircloudy.errors import ConnectionFailed

from .http_client_models import HttpResponse

logger = logging.getLogger(__name__)


async def create_headers(
    host: str, additional_headers: dict[str, str] | None = None, token_supplier: TokenSupplier | None = None
) -> dict[str, str]:
    headers = {
        "Accept": "application/json",
        "Content-Type": "application/json; charset=UTF-8",
        "Host": host,
        "User-Agent": "okhttp/4.2.2",
    }

    if token_supplier is not None:
        headers["Authorization"] = f"Bearer {await token_supplier()}"

    if additional_headers is not None:
        headers.update(additional_headers)

    return headers


async def perform_request(
    method: Literal["GET", "POST", "PUT"],
    url: str,
    additional_headers: dict[str, str] | None = None,
    body: object | None = None,
    token_supplier: TokenSupplier | None = None,
    do_not_raise_exception_on: tuple[int, ...] = (200,),
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> HttpResponse:
    logger.debug("Perform %s %s on %s:%d, %s", method, url, host, port, body)
    try:
        async with (
            aiohttp.ClientSession(f"https://{host}:{port}", connector=TCPConnector(ssl=SSL_CONTEXT)) as session,
            session.request(
                method, url, json=body, headers=await create_headers(host, additional_headers, token_supplier)
            ) as response_http,
        ):
            response_status = response_http.status
            response_body = await response_http.text()
            logger.debug("Response status=%d body=%s", response_status, response_body)

            if response_status not in do_not_raise_exception_on:
                raise Exception(f"Call failed (status={response_status} body={response_body})")

            return HttpResponse(response_status, response_body)
    except aiohttp.client_exceptions.ClientConnectorError as e:
        raise ConnectionFailed(f"Failed to connect to host: {host}") from e



from __future__ import annotations

import asyncio
import logging
import traceback
from asyncio import Task

from aircloudy.api.rac import get_commands_state
from aircloudy.api.rac_models import CommandResponse
from aircloudy.contants import DEFAULT_REST_API_HOST, ApiCommandState, TokenSupplier

logger = logging.getLogger(__name__)


class CommandState:
    _origin: CommandResponse
    _event: asyncio.Event
    _state: ApiCommandState | None

    _state_lock: asyncio.Lock

    def __init__(self, command: CommandResponse) -> None:
        self._command = command
        self._event = asyncio.Event()
        self._state = None
        self._state_lock = asyncio.Lock()

    @property
    def id(self) -> str:
        return self._command.commandId

    @property
    def command(self) -> CommandResponse:
        return self._command

    async def get_state(self) -> ApiCommandState | None:
        async with self._state_lock:
            return self._state

    async def is_terminated(self) -> bool:
        state = await self.get_state()
        return state is None or state == "DONE"

    async def wait_done(self) -> None:
        await self._event.wait()

    def __repr__(self) -> str:
        return (
            f"CommandStatus("
            f"command_id={self._command.commandId}, "
            f"state={self._state}, "
            f"event_is_set={self._event.is_set()})"
        )

    async def set_state(self, state: ApiCommandState) -> None:
        async with self._state_lock:
            self._state = state
            if self._state == "DONE":
                self._event.set()


class CommandStateMonitor:
    _token_supplier: TokenSupplier
    _update_interval: int
    _api_host: str
    _port: int

    _commands: dict[str, CommandState]
    _task_fetch_command_status: Task | None

    _lock: asyncio.Lock

    def __init__(
        self,
        token_supplier: TokenSupplier,
        update_interval: int = 2,
        host: str = DEFAULT_REST_API_HOST,
        port: int = 443,
    ) -> None:
        self._token_supplier = token_supplier
        self._update_interval = update_interval
        self._api_host = host
        self._port = port

        self._lock = asyncio.Lock()
        self._commands = {}
        self._task_fetch_command_status = None

    async def watch_command(self, command: CommandResponse) -> CommandState:
        async with self._lock:
            logger.debug("Add command watch for command %s", command)
            command_status = CommandState(command)
            self._commands[command.commandId] = command_status
            if self._task_fetch_command_status is None or self._task_fetch_command_status.done():
                self._task_fetch_command_status = asyncio.create_task(self._fetch_command_status_loop())
            return command_status

    async def _fetch_command_status_loop(self) -> None:
        logger.debug("Start fetch_command_status_loop")
        try:
            while True:
                async with self._lock:
                    if len(self._commands) == 0:
                        break

                    commands_to_watch = [command_status.command for command_status in self._commands.values()]
                try:
                    commands_state = await get_commands_state(
                        self._token_supplier, commands_to_watch, self._api_host, self._port
                    )

                    async with self._lock:
                        await asyncio.gather(
                            *[self._commands[rac_id].set_state(state) for rac_id, state in commands_state.items()]
                        )
                        self._commands = {
                            cmd.id: cmd for cmd in self._commands.values() if not await cmd.is_terminated()
                        }

                    await asyncio.sleep(self._update_interval)
                except Exception as e:
                    logger.error("Unexpected error while fetching status : %s", traceback.format_exc())
                    raise e
        finally:
            logger.debug("Finish fetch_command_status_loop")



from __future__ import annotations

import asyncio
import datetime
from asyncio import Future
from typing import Awaitable, TypeVar

T = TypeVar("T")


def awaitable(value: T) -> Awaitable[T]:
    f: Future[T] = asyncio.Future()
    f.set_result(value)
    return f


def to_int(value: str | None) -> int | None:
    return int(value) if value else None


def current_task_is_running() -> bool:
    task = asyncio.current_task()
    if task is None:
        return False
    return not task.cancelling() and not task.cancelled()


def utc_datetime_from_millis(millis: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(millis / 1000.0, datetime.timezone.utc)



from __future__ import annotations

import asyncio
import datetime
import logging
import traceback
from dataclasses import dataclass
from typing import Awaitable, Callable

from .api.rac_models import InteriorUnitUserState
from .contants import ApiCommandState, FanSpeed, FanSwing, OperatingMode, Power
from .errors import CommandFailedException, InvalidArgumentException, UnitIsOfflineException
from .interior_unit_base import InteriorUnitBase
from .interior_unit_changes import InteriorUnitChanges

logger = logging.getLogger(__name__)


@dataclass
class NextState:
    _command: InteriorUnitUserState
    _created_at: datetime.datetime

    def __init__(self, command: InteriorUnitUserState) -> None:
        self._command = command
        self._created_at = datetime.datetime.now(datetime.UTC)

    @property
    def command(self) -> InteriorUnitUserState:
        return self._command

    @property
    def created_at(self) -> datetime.datetime:
        return self._created_at

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NextState):
            return False
        return self.command == other.command


@dataclass
class InteriorUnit:
    _send_command_and_wait_ack: Callable[[InteriorUnitUserState], Awaitable[ApiCommandState | None]]

    _id: int
    _name: str
    _room_temperature: float
    _relative_temperature: float
    _updated_at: datetime.datetime
    _online: bool
    _online_updated_at: datetime.datetime
    _vendor: str
    _model_id: str
    _user_state: InteriorUnitUserState

    on_changes: Callable[[InteriorUnitChanges], None] | None = None

    _state_lock: asyncio.Lock = asyncio.Lock()
    _next_state: NextState | None = None
    _state_updater: asyncio.Task | None = None

    def __init__(
        self,
        send_command_and_wait_ack: Callable[[InteriorUnitUserState], Awaitable[None]],
        base: InteriorUnitBase,
    ) -> None:
        self._send_command_and_wait_ack = send_command_and_wait_ack
        self._id = base.rac_id
        self._name = base.name
        self._room_temperature = base.room_temperature
        self._relative_temperature = base.relative_temperature
        self._updated_at = base.updated_at
        self._online = base.online
        self._online_updated_at = base.online_updated_at
        self._vendor = base.vendor
        self._model_id = base.model_id
        self._user_state = base.user_state

    def update(self, base: InteriorUnitBase) -> InteriorUnitChanges:
        if base.rac_id != self.id:
            raise InvalidArgumentException("Update must come from the same id")
        changes = InteriorUnitChanges(
            (self._name, base.name) if self._name != base.name else None,
            (
                (self._room_temperature, base.room_temperature)
                if self._room_temperature != base.room_temperature
                else None
            ),
            (
                (self._relative_temperature, base.relative_temperature)
                if self._relative_temperature != base.relative_temperature
                else None
            ),
            (self._updated_at, base.updated_at) if self._updated_at != base.updated_at else None,
            (self._online, base.online) if self._online != base.online else None,
            (
                (self._online_updated_at, base.online_updated_at)
                if self._online_updated_at != base.online_updated_at
                else None
            ),
            (self._vendor, base.vendor) if self._vendor != base.vendor else None,
            (self._model_id, base.model_id) if self._model_id != base.model_id else None,
            (self._user_state.power, base.power) if self._user_state._power != base.power else None,
            (
                (self._user_state.operating_mode, base.operating_mode)
                if self._user_state._operating_mode != base.operating_mode
                else None
            ),
            (
                (self._user_state.requested_temperature, base.requested_temperature)
                if self._user_state._requested_temperature != base.requested_temperature
                else None
            ),
            (self._user_state.humidity, base.humidity) if self._user_state._humidity != base.humidity else None,
            (self._user_state.fan_speed, base.fan_speed) if self._user_state._fan_speed != base.fan_speed else None,
            (self._user_state.fan_swing, base.fan_swing) if self._user_state._fan_swing != base.fan_swing else None,
        )

        self._name = base.name
        self._room_temperature = base.room_temperature
        self._relative_temperature = base.relative_temperature
        self._updated_at = base.updated_at
        self._online = base.online
        self._online_updated_at = base.online_updated_at
        self._vendor = base.vendor
        self._model_id = base.model_id
        self._user_state = base.user_state

        if self.on_changes is not None:
            self.on_changes(changes)

        return changes

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def room_temperature(self) -> float:
        return self._room_temperature

    @property
    def relative_temperature(self) -> float:
        return self._relative_temperature

    @property
    def updated_at(self) -> datetime.datetime:
        return self._updated_at

    @property
    def online(self) -> bool:
        return self._online

    @property
    def online_updated_at(self) -> datetime.datetime:
        return self._online_updated_at

    @property
    def vendor(self) -> str:
        return self._vendor

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def power(self) -> Power:
        return self._user_state._power

    @property
    def operating_mode(self) -> OperatingMode:
        return self._user_state._operating_mode

    @property
    def requested_temperature(self) -> float:
        return self._user_state._requested_temperature

    @property
    def humidity(self) -> int:
        return self._user_state._humidity

    @property
    def fan_speed(self) -> FanSpeed:
        return self._user_state._fan_speed

    @property
    def fan_swing(self) -> FanSwing:
        return self._user_state._fan_swing

    async def send_command(
        self,
        power: Power | None = None,
        mode: OperatingMode | None = None,
        requested_temperature: float | None = None,
        humidity: int | None = None,
        fan_speed: FanSpeed | None = None,
        fan_swing: FanSwing | None = None,
    ) -> None:
        if not self._online:
            raise UnitIsOfflineException

        async with self._state_lock:
            base_state = (
                self._next_state.command
                if self._next_state is not None and self._next_state.created_at > self._updated_at
                else self._user_state
            )
            self._next_state = NextState(
                base_state.copy(power, mode, requested_temperature, humidity, fan_speed, fan_swing)
            )
            if self._state_updater is None or self._state_updater.done():
                self._state_updater = asyncio.create_task(self._update_state())

    async def _update_state(self) -> None:
        last_state: NextState | None = None
        while True:
            async with self._state_lock:
                new_state = self._next_state
                if new_state is None or new_state == last_state:
                    break

            try:
                await self._send_command_and_wait_ack(new_state.command)
                async with self._state_lock:
                    self._user_state = new_state.command
                    self._updated_at = new_state.created_at
            except CommandFailedException:
                logger.warning("Failed to acknowledge command execution: %s", traceback.format_exc())

            last_state = new_state

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InteriorUnit):
            return False

        return (
            self._id == other._id
            and self._name == other._name
            and self._room_temperature == other._room_temperature
            and self._relative_temperature == other._relative_temperature
            and self._updated_at == other._updated_at
            and self._online == other._online
            and self._online_updated_at == other._online_updated_at
            and self._vendor == other._vendor
            and self._model_id == other._model_id
            and self._user_state == other._user_state
        )



from __future__ import annotations

import json
import logging

from aircloudy.contants import DEFAULT_REST_API_HOST, ApiCommandState, Power, TokenSupplier

from ..errors import TooManyRequestsException
from ..interior_unit_base import InteriorUnitBase
from ..utils import utc_datetime_from_millis
from .http_client import perform_request
from .rac_models import CommandResponse, InteriorUnitUserState, PowerAllResponse

logger = logging.getLogger(__name__)


async def get_interior_units(
    token_supplier: TokenSupplier, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> list[InteriorUnitBase]:
    response = await perform_request(
        "GET", f"/rac/ownership/groups/{family_id}/idu-list", token_supplier=token_supplier, host=host, port=port
    )

    if response.status != 200:
        raise Exception(f"Call failed (status={response.status} body={response.body}")

    return [
        InteriorUnitBase(
            d["id"],
            d["name"],
            d["roomTemperature"],
            d["relativeTemperature"],
            utc_datetime_from_millis(d["updatedAt"]),
            d["online"],
            utc_datetime_from_millis(d["lastOnlineUpdatedAt"]),
            d["model"],
            str(d["racTypeId"]),
            d["serialNumber"],
            d["vendorThingId"],
            d["scheduleType"],
            d["power"],
            d["mode"],
            d["iduTemperature"],
            d["humidity"],
            d["fanSpeed"],
            d["fanSwing"],
        )
        for d in response.body_as_json
    ]


async def get_commands_state(
    token_supplier: TokenSupplier, commands: list[CommandResponse], host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> dict[str, ApiCommandState]:
    response = await perform_request(
        "POST",
        "/rac/status/command",
        body=[c.__dict__ for c in commands],
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    return {item["commandId"]: item["status"] for item in response.body_as_json}


async def send_command(
    token_supplier: TokenSupplier,
    family_id: int,
    command: InteriorUnitUserState,
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> CommandResponse:
    """Send command to change interior unit (like a remote control)

    :raises:
        TooManyRequestsException: If previous command is still in progress
    """
    body = command.to_api()
    logger.info("Configure interior unit family_id=%s : %s", family_id, body)
    response = await perform_request(
        "PUT",
        f"/rac/basic-idu-control/general-control-command/{command.rac_id}?familyId={family_id}",
        body=body,
        do_not_raise_exception_on=(200, 429),
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    if response.status == 429:
        raise TooManyRequestsException(response.body)

    return CommandResponse(response.body_as_json)


async def request_refresh_interior_unit_state(
    token_supplier: TokenSupplier, rac_id: int, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> None:
    logger.debug("Request refresh interior unit state for rac id=%s, family_id=%s", rac_id, family_id)
    await perform_request(
        "PUT", f"/rac/status/{rac_id}?familyId={family_id}", token_supplier=token_supplier, host=host, port=port
    )


async def set_power(
    token_supplier: TokenSupplier, rac_id: str, power: Power, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> None:
    logger.debug("Set power rac_id=%s, power=%s", rac_id, power)
    await perform_request(
        "PUT",
        f"/rac/basic-idu-control/switch-on-off/{rac_id}",
        body={
            "power": power,
        },
        token_supplier=token_supplier,
        host=host,
        port=port,
    )


async def set_power_all(
    token_supplier: TokenSupplier,
    family_id: str,
    power: Power,
    interior_units_state: list[InteriorUnitUserState],
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> PowerAllResponse:
    match power:
        case "ON":
            url = f"/rac/manage-idu/groups/{family_id}/idu/start"
        case "OFF":
            url = f"/rac/manage-idu/groups/{family_id}/idu/stop"
        case _:
            raise Exception(f"Unknown power value {power}")

    units = [iu.copy(power=power).to_api() for iu in interior_units_state]

    logger.debug("Set power all power=%s for %s", power, units)
    response = await perform_request(
        "PUT",
        url,
        body=json.dumps(units),
        do_not_raise_exception_on=(200, 207),
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    return PowerAllResponse(response.body_as_json)



from __future__ import annotations

import datetime
from dataclasses import dataclass

from .contants import FanSpeed, FanSwing, OperatingMode, Power


@dataclass
class InteriorUnitChanges:
    name: tuple[str, str] | None
    room_temperature: tuple[float, float] | None
    relative_temperature: tuple[float, float] | None
    updated_at: tuple[datetime.datetime, datetime.datetime] | None
    online: tuple[bool, bool] | None
    online_updated_at: tuple[datetime.datetime, datetime.datetime] | None
    vendor: tuple[str, str] | None
    model_id: tuple[str, str] | None
    power: tuple[Power, Power] | None
    operating_mode: tuple[OperatingMode, OperatingMode] | None
    requested_temperature: tuple[float, float] | None
    humidity: tuple[int, int] | None
    fan_speed: tuple[FanSpeed, FanSpeed] | None
    fan_swing: tuple[FanSwing, FanSwing] | None

    @property
    def has_changes(self) -> bool:
        return (
            self.name is not None
            or self.room_temperature is not None
            or self.relative_temperature is not None
            or self.updated_at is not None
            or self.online is not None
            or self.online_updated_at is not None
            or self.vendor is not None
            or self.model_id is not None
            or self.power is not None
            or self.operating_mode is not None
            or self.requested_temperature is not None
            or self.humidity is not None
            or self.fan_speed is not None
            or self.fan_swing is not None
        )

    def __repr__(self) -> str:
        changes_as_string: list[str] = []
        if self.name is not None:
            changes_as_string.append(f"name={self.name[0]}->{self.name[1]}")
        if self.room_temperature is not None:
            changes_as_string.append(f"room_temperature={self.room_temperature[0]}->{self.room_temperature[1]}")
        if self.relative_temperature is not None:
            changes_as_string.append(
                f"relative_temperature={self.relative_temperature[0]}->{self.relative_temperature[1]}"
            )
        if self.updated_at is not None:
            changes_as_string.append(f"updated_at={self.updated_at[0]}->{self.updated_at[1]}")
        if self.online is not None:
            changes_as_string.append(f"online={self.online[0]}->{self.online[1]}")
        if self.online_updated_at is not None:
            changes_as_string.append(f"online_updated_at={self.online_updated_at[0]}->{self.online_updated_at[1]}")
        if self.vendor is not None:
            changes_as_string.append(f"vendor={self.vendor[0]}->{self.vendor[1]}")
        if self.model_id is not None:
            changes_as_string.append(f"model_id={self.model_id[0]}->{self.model_id[1]}")
        if self.power is not None:
            changes_as_string.append(f"power={self.power[0]}->{self.power[1]}")
        if self.operating_mode is not None:
            changes_as_string.append(f"operating_mode={self.operating_mode[0]}->{self.operating_mode[1]}")
        if self.requested_temperature is not None:
            changes_as_string.append(
                f"requested_temperature={self.requested_temperature[0]}->{self.requested_temperature[1]}"
            )
        if self.humidity is not None:
            changes_as_string.append(f"humidity={self.humidity[0]}->{self.humidity[1]}")
        if self.fan_speed is not None:
            changes_as_string.append(f"fan_speed={self.fan_speed[0]}->{self.fan_speed[1]}")
        if self.fan_swing is not None:
            changes_as_string.append(f"fan_swing={self.fan_swing[0]}->{self.fan_swing[1]}")

        return f"InteriorUnitChanges({', '.join(changes_as_string)})"
